@@ -1,6 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
  *
  * Copyright (C) 2004-2008 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2008-2011 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,20 +17,19 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
- * Authors: William Jon McCann <mccann@jhu.edu>
- *
  */
 
 #include "config.h"
 
+#include <sys/time.h>
 #include <sys/types.h>
 #include <errno.h>
 #include <sys/wait.h>
 #include <string.h>
 
-#include <gdk/gdkx.h>
-#include <gdk/gdkkeysyms.h>
+#include <glib/gi18n.h>
 #include <gtk/gtk.h>
+#include <gtk/gtkx.h>
 
 #include "gs-window.h"
 #include "gs-marshal.h"
@@ -77,6 +77,9 @@ struct GSWindowPrivate
         char      *status_message;
 
         GtkWidget *vbox;
+        GtkWidget *panel;
+        GtkWidget *clock;
+        GtkWidget *name_label;
         GtkWidget *drawing_area;
         GtkWidget *lock_box;
         GtkWidget *lock_socket;
@@ -84,7 +87,7 @@ struct GSWindowPrivate
         GtkWidget *info_bar;
         GtkWidget *info_content;
 
-        GdkPixmap *background_pixmap;
+        cairo_surface_t *background_surface;
 
         guint      popup_dialog_idle_id;
 
@@ -94,6 +97,7 @@ struct GSWindowPrivate
 
         guint      watchdog_timer_id;
         guint      info_bar_timer_id;
+        guint      clock_update_id;
 
         gint       lock_pid;
         gint       lock_watch_id;
@@ -140,29 +144,16 @@ static guint           signals [LAST_SIGNAL] = { 0, };
 
 G_DEFINE_TYPE (GSWindow, gs_window, GTK_TYPE_WINDOW)
 
+static void queue_clock_update (GSWindow *window);
+
 static void
 set_invisible_cursor (GdkWindow *window,
                       gboolean   invisible)
 {
-        GdkBitmap *empty_bitmap;
         GdkCursor *cursor = NULL;
-        GdkColor   useless;
-        char       invisible_cursor_bits [] = { 0x0 };
 
         if (invisible) {
-                useless.red = useless.green = useless.blue = 0;
-                useless.pixel = 0;
-
-                empty_bitmap = gdk_bitmap_create_from_data (window,
-                                                            invisible_cursor_bits,
-                                                            1, 1);
-
-                cursor = gdk_cursor_new_from_pixmap (empty_bitmap,
-                                                     empty_bitmap,
-                                                     &useless,
-                                                     &useless, 0, 0);
-
-                g_object_unref (empty_bitmap);
+                cursor = gdk_cursor_new (GDK_BLANK_CURSOR);
         }
 
         gdk_window_set_cursor (window, cursor);
@@ -189,10 +180,10 @@ gs_window_override_user_time (GSWindow *window)
                  * NOTE: Last resort for D-BUS or other non-interactive
                  *       openings.  Causes roundtrip to server.  Lame.
                  */
-                ev_time = gdk_x11_get_server_time (GTK_WIDGET (window)->window);
+                ev_time = gdk_x11_get_server_time (gtk_widget_get_window (GTK_WIDGET (window)));
         }
 
-        gdk_x11_window_set_user_time (GTK_WIDGET (window)->window, ev_time);
+        gdk_x11_window_set_user_time (gtk_widget_get_window (GTK_WIDGET (window)), ev_time);
 }
 
 static void
@@ -222,165 +213,57 @@ force_no_pixmap_background (GtkWidget *widget)
 }
 
 static void
-clear_children (Window window)
+gs_window_reset_background_surface (GSWindow *window)
 {
-        Window            root;
-        Window            parent;
-        Window           *children;
-        unsigned int      n_children;
-        int               status;
-
-        children = NULL;
-        status = XQueryTree (GDK_DISPLAY (), window, &root, &parent, &children, &n_children);
-
-        if (status == 0) {
-                if (children) {
-                        XFree (children);
-                }
-                return;
-        }
-
-        if (children) {
-                while (n_children) {
-                        Window child;
-
-                        child = children [--n_children];
-
-                        XClearWindow (GDK_DISPLAY (), child);
-                        clear_children (child);
-                }
-
-                XFree (children);
-        }
-}
-
-static void
-widget_clear_all_children (GtkWidget *widget)
-{
-        GdkWindow *w;
-
-        gs_debug ("Clearing all child windows");
-
-        gdk_error_trap_push ();
-
-        w = widget->window;
-
-        clear_children (GDK_WINDOW_XID (w));
-
-        gdk_display_sync (gtk_widget_get_display (widget));
-        gdk_error_trap_pop ();
+        cairo_pattern_t *pattern;
+        pattern = cairo_pattern_create_for_surface (window->priv->background_surface);
+        gdk_window_set_background_pattern (gtk_widget_get_window (GTK_WIDGET (window)),
+                                           pattern);
+        cairo_pattern_destroy (pattern);
+        gtk_widget_queue_draw (GTK_WIDGET (window));
 }
 
 void
-gs_window_set_background_pixmap (GSWindow  *window,
-                                 GdkPixmap *pixmap)
+gs_window_set_background_surface (GSWindow        *window,
+                                  cairo_surface_t *surface)
 {
         g_return_if_fail (GS_IS_WINDOW (window));
 
-        if (window->priv->background_pixmap != NULL) {
-                g_object_unref (window->priv->background_pixmap);
+        if (window->priv->background_surface != NULL) {
+                cairo_surface_destroy (window->priv->background_surface);
         }
 
-        if (pixmap != NULL) {
-                window->priv->background_pixmap = g_object_ref (pixmap);
-                gdk_window_set_back_pixmap (GTK_WIDGET (window)->window,
-                                            pixmap,
-                                            FALSE);
+        if (surface != NULL) {
+                window->priv->background_surface = cairo_surface_reference (surface);
+                gs_window_reset_background_surface (window);
         }
 }
 
 static void
-gs_window_clear_to_background_pixmap (GSWindow *window)
+gs_window_clear_to_background_surface (GSWindow *window)
 {
-        GtkStateType state;
-        GtkStyle    *style;
-
         g_return_if_fail (GS_IS_WINDOW (window));
 
-        if (! GTK_WIDGET_VISIBLE (GTK_WIDGET (window))) {
+        if (!gtk_widget_get_visible (GTK_WIDGET (window))) {
                 return;
         }
 
-        if (window->priv->background_pixmap == NULL) {
-                /* don't allow null pixmaps */
+        if (window->priv->background_surface == NULL) {
                 return;
         }
 
         gs_debug ("Clearing window to background pixmap");
-
-        style = gtk_style_copy (GTK_WIDGET (window)->style);
-
-        state = (GtkStateType) 0;
-        while (state < (GtkStateType) G_N_ELEMENTS (GTK_WIDGET (window)->style->bg_pixmap)) {
-
-                if (style->bg_pixmap[state] != NULL) {
-                        g_object_unref (style->bg_pixmap[state]);
-                }
-
-                style->bg_pixmap[state] = g_object_ref (window->priv->background_pixmap);
-                state++;
-        }
-
-        gtk_widget_set_style (GTK_WIDGET (window), style);
-        g_object_unref (style);
-
-        if (window->priv->background_pixmap != NULL) {
-                gdk_window_set_back_pixmap (GTK_WIDGET (window)->window,
-                                            window->priv->background_pixmap,
-                                            FALSE);
-        }
-
-        gdk_window_clear (GTK_WIDGET (window)->window);
-
-        gdk_flush ();
+        gs_window_reset_background_surface (window);
 }
 
 static void
 clear_widget (GtkWidget *widget)
 {
-        GdkColor     color = { 0, 0x0000, 0x0000, 0x0000 };
-        GdkColormap *colormap;
-        GtkStateType state;
-        GtkStyle    *style;
+        GdkColor color = { 0, 0x0000, 0x0000, 0x0000 };
 
-        if (! GTK_WIDGET_VISIBLE (widget)) {
-                return;
-        }
-
-        gs_debug ("Clearing widget");
-
-        state = (GtkStateType) 0;
-        while (state < (GtkStateType) G_N_ELEMENTS (widget->style->bg)) {
-                gtk_widget_modify_bg (widget, state, &color);
-                state++;
-        }
-
-        style = gtk_style_copy (widget->style);
-
-        state = (GtkStateType) 0;
-        while (state < (GtkStateType) G_N_ELEMENTS (widget->style->bg_pixmap)) {
-
-                if (style->bg_pixmap[state] != NULL) {
-                        g_object_unref (style->bg_pixmap[state]);
-                }
-
-                style->bg_pixmap[state] = NULL;
-                state++;
-        }
-
-        colormap = gdk_drawable_get_colormap (widget->window);
-        gdk_colormap_alloc_color (colormap, &color, FALSE, TRUE);
-        gdk_window_set_background (widget->window, &color);
-
-        gtk_widget_set_style (widget, style);
-        g_object_unref (style);
-
-        gdk_window_clear (widget->window);
-
-        /* If a screensaver theme adds child windows we need to clear them too */
-        widget_clear_all_children (widget);
-
-        gdk_flush ();
+        gdk_window_set_background (gtk_widget_get_window (widget),
+                                   &color);
+        gtk_widget_queue_draw (GTK_WIDGET (widget));
 }
 
 void
@@ -392,19 +275,24 @@ gs_window_clear (GSWindow *window)
         clear_widget (window->priv->drawing_area);
 }
 
-static GdkRegion *
+static cairo_region_t *
 get_outside_region (GSWindow *window)
 {
-        int        i;
-        GdkRegion *region;
+        int             i;
+        cairo_region_t *region;
 
-        region = gdk_region_new ();
+        region = cairo_region_create ();
         for (i = 0; i < window->priv->monitor; i++) {
                 GdkRectangle geometry;
+                cairo_rectangle_int_t rectangle;
 
-                gdk_screen_get_monitor_geometry (GTK_WINDOW (window)->screen,
+                gdk_screen_get_monitor_geometry (gtk_window_get_screen (GTK_WINDOW (window)),
                                                    i, &geometry);
-                gdk_region_union_with_rect (region, &geometry);
+                rectangle.x = geometry.x;
+                rectangle.y = geometry.y;
+                rectangle.width = geometry.width;
+                rectangle.height = geometry.height;
+                cairo_region_union_rectangle (region, &rectangle);
         }
 
         return region;
@@ -413,13 +301,13 @@ get_outside_region (GSWindow *window)
 static void
 update_geometry (GSWindow *window)
 {
-        GdkRectangle geometry;
-        GdkRegion   *outside_region;
-        GdkRegion   *monitor_region;
+        GdkRectangle    geometry;
+        cairo_region_t *outside_region;
+        cairo_region_t *monitor_region;
 
         outside_region = get_outside_region (window);
 
-        gdk_screen_get_monitor_geometry (GTK_WINDOW (window)->screen,
+        gdk_screen_get_monitor_geometry (gtk_window_get_screen (GTK_WINDOW (window)),
                                          window->priv->monitor,
                                          &geometry);
         gs_debug ("got geometry for monitor %d: x=%d y=%d w=%d h=%d",
@@ -428,12 +316,12 @@ update_geometry (GSWindow *window)
                   geometry.y,
                   geometry.width,
                   geometry.height);
-        monitor_region = gdk_region_rectangle (&geometry);
-        gdk_region_subtract (monitor_region, outside_region);
-        gdk_region_destroy (outside_region);
+        monitor_region = cairo_region_create_rectangle ((const cairo_rectangle_int_t *)&geometry);
+        cairo_region_subtract (monitor_region, outside_region);
+        cairo_region_destroy (outside_region);
 
-        gdk_region_get_clipbox (monitor_region, &geometry);
-        gdk_region_destroy (monitor_region);
+        cairo_region_get_extents (monitor_region, (cairo_rectangle_int_t *)&geometry);
+        cairo_region_destroy (monitor_region);
 
         gs_debug ("using geometry for monitor %d: x=%d y=%d w=%d h=%d",
                   window->priv->monitor,
@@ -466,7 +354,7 @@ gs_window_move_resize_window (GSWindow *window,
 
         widget = GTK_WIDGET (window);
 
-        g_assert (GTK_WIDGET_REALIZED (widget));
+        g_assert (gtk_widget_get_realized (widget));
 
         gs_debug ("Move and/or resize window on monitor %d: x=%d y=%d w=%d h=%d",
                   window->priv->monitor,
@@ -476,17 +364,17 @@ gs_window_move_resize_window (GSWindow *window,
                   window->priv->geometry.height);
 
         if (move && resize) {
-                gdk_window_move_resize (widget->window,
+                gdk_window_move_resize (gtk_widget_get_window (widget),
                                         window->priv->geometry.x,
                                         window->priv->geometry.y,
                                         window->priv->geometry.width,
                                         window->priv->geometry.height);
         } else if (move) {
-                gdk_window_move (widget->window,
+                gdk_window_move (gtk_widget_get_window (widget),
                                  window->priv->geometry.x,
                                  window->priv->geometry.y);
         } else if (resize) {
-                gdk_window_resize (widget->window,
+                gdk_window_resize (gtk_widget_get_window (widget),
                                    window->priv->geometry.width,
                                    window->priv->geometry.height);
         }
@@ -615,7 +503,7 @@ get_best_visual_for_screen (GdkScreen *screen)
                         VisualID      visual_id;
 
                         visual_id = (VisualID) v;
-                        visual = gdkx_visual_get (visual_id);
+                        visual = gdk_x11_screen_lookup_visual (screen, visual_id);
 
                         gs_debug ("Found best GL visual for screen %d: 0x%x",
                                   gdk_screen_get_number (screen),
@@ -626,45 +514,27 @@ get_best_visual_for_screen (GdkScreen *screen)
         g_free (std_output);
         g_free (command);
 
-        return visual;
-}
-
-static GdkColormap *
-get_best_colormap_for_screen (GdkScreen *screen)
-{
-        GdkColormap *colormap;
-        GdkVisual   *visual;
-
-        g_return_val_if_fail (screen != NULL, NULL);
-
-        visual = get_best_visual_for_screen (screen);
-
-        colormap = NULL;
-        if (visual != NULL) {
-                colormap = gdk_colormap_new (visual, FALSE);
-        }
-
-        return colormap;
+        return g_object_ref (visual);
 }
 
 static void
-widget_set_best_colormap (GtkWidget *widget)
+widget_set_best_visual (GtkWidget *widget)
 {
-        GdkColormap *colormap;
+        GdkVisual *visual;
 
         g_return_if_fail (widget != NULL);
 
-        colormap = get_best_colormap_for_screen (gtk_widget_get_screen (widget));
-        if (colormap != NULL) {
-                gtk_widget_set_colormap (widget, colormap);
-                g_object_unref (colormap);
+        visual = get_best_visual_for_screen (gtk_widget_get_screen (widget));
+        if (visual != NULL) {
+                gtk_widget_set_visual (widget, visual);
+                g_object_unref (visual);
         }
 }
 
 static void
 gs_window_real_realize (GtkWidget *widget)
 {
-        widget_set_best_colormap (widget);
+        widget_set_best_visual (widget);
 
         if (GTK_WIDGET_CLASS (gs_window_parent_class)->realize) {
                 GTK_WIDGET_CLASS (gs_window_parent_class)->realize (widget);
@@ -687,7 +557,7 @@ watchdog_timer (GSWindow *window)
 {
         GtkWidget *widget = GTK_WIDGET (window);
 
-        gdk_window_focus (widget->window, GDK_CURRENT_TIME);
+        gdk_window_focus (gtk_widget_get_window (widget), GDK_CURRENT_TIME);
 
         return TRUE;
 }
@@ -748,7 +618,7 @@ gs_window_raise (GSWindow *window)
 
         gs_debug ("Raising screensaver window");
 
-        win = GTK_WIDGET (window)->window;
+        win = gtk_widget_get_window (GTK_WIDGET (window));
 
         gdk_window_raise (win);
 }
@@ -761,7 +631,7 @@ x11_window_is_ours (Window window)
 
         ret = FALSE;
 
-        gwindow = gdk_window_lookup (window);
+        gwindow = gdk_x11_window_lookup_for_display (gdk_display_get_default (), window);
         if (gwindow && (window != GDK_ROOT_WINDOW ())) {
                 ret = TRUE;
         }
@@ -773,7 +643,7 @@ x11_window_is_ours (Window window)
 static void
 unshape_window (GSWindow *window)
 {
-        gdk_window_shape_combine_region (GTK_WIDGET (window)->window,
+        gdk_window_shape_combine_region (gtk_widget_get_window (GTK_WIDGET (window)),
                                          NULL,
                                          0,
                                          0);
@@ -849,13 +719,12 @@ select_popup_events (void)
         gdk_error_trap_push ();
 
         memset (&attr, 0, sizeof (attr));
-        XGetWindowAttributes (GDK_DISPLAY (), GDK_ROOT_WINDOW (), &attr);
+        XGetWindowAttributes (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), GDK_ROOT_WINDOW (), &attr);
 
         events = SubstructureNotifyMask | attr.your_event_mask;
-        XSelectInput (GDK_DISPLAY (), GDK_ROOT_WINDOW (), events);
+        XSelectInput (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), GDK_ROOT_WINDOW (), events);
 
-        gdk_display_sync (gdk_display_get_default ());
-        gdk_error_trap_pop ();
+        gdk_error_trap_pop_ignored ();
 }
 
 static void
@@ -867,13 +736,12 @@ window_select_shape_events (GSWindow *window)
 
         gdk_error_trap_push ();
 
-        if (XShapeQueryExtension (GDK_DISPLAY (), &window->priv->shape_event_base, &shape_error_base)) {
+        if (XShapeQueryExtension (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), &window->priv->shape_event_base, &shape_error_base)) {
                 events = ShapeNotifyMask;
-                XShapeSelectInput (GDK_DISPLAY (), GDK_WINDOW_XID (GTK_WIDGET (window)->window), events);
+                XShapeSelectInput (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), GDK_WINDOW_XID (gtk_widget_get_window (GTK_WIDGET (window))), events);
         }
 
-        gdk_display_sync (gdk_display_get_default ());
-        gdk_error_trap_pop ();
+        gdk_error_trap_pop_ignored ();
 #endif
 }
 
@@ -888,7 +756,7 @@ gs_window_real_show (GtkWidget *widget)
 
         gs_window_clear (GS_WINDOW (widget));
 
-        set_invisible_cursor (widget->window, TRUE);
+        set_invisible_cursor (gtk_widget_get_window (widget), TRUE);
 
         window = GS_WINDOW (widget);
         if (window->priv->timer) {
@@ -1033,7 +901,7 @@ gs_window_get_gdk_window (GSWindow *window)
 {
         g_return_val_if_fail (GS_IS_WINDOW (window), NULL);
 
-        return GTK_WIDGET (window)->window;
+        return gtk_widget_get_window (GTK_WIDGET (window));
 }
 
 GtkWidget *
@@ -1098,6 +966,7 @@ spawn_on_window (GSWindow *window,
 {
         int         argc;
         char      **argv;
+        char      **envp;
         GError     *error;
         gboolean    result;
         GIOChannel *channel;
@@ -1113,19 +982,20 @@ spawn_on_window (GSWindow *window,
                 return FALSE;
         }
 
+        envp = spawn_make_environment_for_screen (gtk_window_get_screen (GTK_WINDOW (window)), NULL);
+
         error = NULL;
-        result = gdk_spawn_on_screen_with_pipes (GTK_WINDOW (window)->screen,
-                                                 NULL,
-                                                 argv,
-                                                 NULL,
-                                                 G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
-                                                 NULL,
-                                                 NULL,
-                                                 &child_pid,
-                                                 NULL,
-                                                 &standard_output,
-                                                 &standard_error,
-                                                 &error);
+        result = g_spawn_async_with_pipes (NULL,
+                                           argv,
+                                           envp,
+                                           G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
+                                           NULL,
+                                           NULL,
+                                           &child_pid,
+                                           NULL,
+                                           &standard_output,
+                                           &standard_error,
+                                           &error);
 
         if (! result) {
                 gs_debug ("Could not start command '%s': %s", command, error->message);
@@ -1168,6 +1038,7 @@ spawn_on_window (GSWindow *window,
         g_io_channel_unref (channel);
 
         g_strfreev (argv);
+        g_strfreev (envp);
 
         return result;
 }
@@ -1552,7 +1423,7 @@ popdown_dialog (GSWindow *window)
         gtk_widget_show (window->priv->drawing_area);
 
         gs_window_clear (window);
-        set_invisible_cursor (GTK_WIDGET (window)->window, TRUE);
+        set_invisible_cursor (gtk_widget_get_window (GTK_WIDGET (window)), TRUE);
 
         window_set_dialog_up (window, FALSE);
 
@@ -1712,9 +1583,9 @@ popup_dialog (GSWindow *window)
 
         gtk_widget_hide (window->priv->drawing_area);
 
-        gs_window_clear_to_background_pixmap (window);
+        gs_window_clear_to_background_surface (window);
 
-        set_invisible_cursor (GTK_WIDGET (window)->window, FALSE);
+        set_invisible_cursor (gtk_widget_get_window (GTK_WIDGET (window)), FALSE);
 
         window->priv->dialog_quit_requested = FALSE;
         window->priv->dialog_shake_in_progress = FALSE;
@@ -1749,7 +1620,7 @@ gs_window_request_unlock (GSWindow *window)
 
         gs_debug ("Requesting unlock");
 
-        if (! GTK_WIDGET_VISIBLE (GTK_WIDGET (window))) {
+        if (! gtk_widget_get_visible (GTK_WIDGET (window))) {
                 gs_debug ("Request unlock but window is not visible!");
                 return;
         }
@@ -1813,7 +1684,7 @@ gs_window_get_screen (GSWindow  *window)
 {
         g_return_val_if_fail (GS_IS_WINDOW (window), NULL);
 
-        return GTK_WINDOW (window)->screen;
+        return gtk_window_get_screen (GTK_WINDOW (window));
 }
 
 void
@@ -2015,10 +1886,10 @@ queue_key_event (GSWindow    *window,
 {
         /* Eat the first return, enter, escape, or space */
         if (window->priv->key_events == NULL
-            && (event->keyval == GDK_Return
-                || event->keyval == GDK_KP_Enter
-                || event->keyval == GDK_Escape
-                || event->keyval == GDK_space)) {
+            && (event->keyval == GDK_KEY_Return
+                || event->keyval == GDK_KEY_KP_Enter
+                || event->keyval == GDK_KEY_Escape
+                || event->keyval == GDK_KEY_space)) {
                 return;
         }
 
@@ -2026,9 +1897,9 @@ queue_key_event (GSWindow    *window,
            something is wrong */
         /* Don't queue keys that may cause focus navigation in the dialog */
         if (g_list_length (window->priv->key_events) < MAX_QUEUED_EVENTS
-            && event->keyval != GDK_Tab
-            && event->keyval != GDK_Up
-            && event->keyval != GDK_Down) {
+            && event->keyval != GDK_KEY_Tab
+            && event->keyval != GDK_KEY_Up
+            && event->keyval != GDK_KEY_Down) {
                 window->priv->key_events = g_list_prepend (window->priv->key_events,
                                                            gdk_event_copy ((GdkEvent *)event));
         }
@@ -2043,7 +1914,7 @@ maybe_handle_activity (GSWindow *window)
 
         /* if we already have a socket then don't bother */
         if (! window->priv->lock_socket
-            && GTK_WIDGET_IS_SENSITIVE (GTK_WIDGET (window))) {
+            && gtk_widget_get_sensitive (GTK_WIDGET (window))) {
                 g_signal_emit (window, signals [ACTIVITY], 0, &handled);
         }
 
@@ -2147,8 +2018,8 @@ gs_window_real_size_request (GtkWidget      *widget,
         window = GS_WINDOW (widget);
         bin = GTK_BIN (widget);
 
-        if (bin->child && GTK_WIDGET_VISIBLE (bin->child)) {
-                gtk_widget_size_request (bin->child, requisition);
+        if (gtk_bin_get_child (bin) && gtk_widget_get_visible (gtk_bin_get_child (bin))) {
+                gtk_widget_size_request (gtk_bin_get_child (bin), requisition);
         }
 
         old_geometry = window->priv->geometry;
@@ -2158,7 +2029,7 @@ gs_window_real_size_request (GtkWidget      *widget,
         requisition->width  = window->priv->geometry.width;
         requisition->height = window->priv->geometry.height;
 
-        if (! GTK_WIDGET_REALIZED (widget)) {
+        if (!gtk_widget_get_realized (widget)) {
                 return;
         }
 
@@ -2241,6 +2112,30 @@ gs_window_real_visibility_notify_event (GtkWidget          *widget,
         return FALSE;
 }
 
+static void
+gs_window_real_get_preferred_width (GtkWidget *widget,
+                               gint      *minimal_width,
+                               gint      *natural_width)
+{
+	GtkRequisition requisition;
+
+	gs_window_real_size_request (widget, &requisition);
+
+	*minimal_width = *natural_width = requisition.width;
+}
+
+static void
+gs_window_real_get_preferred_height (GtkWidget *widget,
+                                gint      *minimal_height,
+                                gint      *natural_height)
+{
+	GtkRequisition requisition;
+
+	gs_window_real_size_request (widget, &requisition);
+
+	*minimal_height = *natural_height = requisition.height;
+}
+
 static void
 gs_window_class_init (GSWindowClass *klass)
 {
@@ -2259,7 +2154,8 @@ gs_window_class_init (GSWindowClass *klass)
         widget_class->motion_notify_event = gs_window_real_motion_notify_event;
         widget_class->button_press_event  = gs_window_real_button_press_event;
         widget_class->scroll_event        = gs_window_real_scroll_event;
-        widget_class->size_request        = gs_window_real_size_request;
+        widget_class->get_preferred_width        = gs_window_real_get_preferred_width;
+        widget_class->get_preferred_height       = gs_window_real_get_preferred_height;
         widget_class->grab_broken_event   = gs_window_real_grab_broken;
         widget_class->visibility_notify_event = gs_window_real_visibility_notify_event;
 
@@ -2372,6 +2268,167 @@ create_info_bar (GSWindow *window)
         gtk_box_pack_end (GTK_BOX (window->priv->vbox), window->priv->info_bar, FALSE, FALSE, 0);
 }
 
+static gboolean
+on_panel_draw (GtkWidget    *widget,
+               cairo_t      *cr,
+               GSWindow     *window)
+{
+        cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
+        cairo_paint (cr);
+
+        return FALSE;
+}
+
+static void
+update_clock (GSWindow *window)
+{
+        const char *clock_format;
+        char *text;
+        char *markup;
+        gboolean use_24 = FALSE;
+        GDateTime *dt;
+
+        /* clock */
+        /* FIXME: set 12/24 hour */
+        if (use_24)
+                clock_format = _("%a %R");
+        else
+                clock_format = _("%a %l:%M %p");
+
+        dt = g_date_time_new_now_local ();
+        text = g_date_time_format (dt, clock_format);
+        markup = g_strdup_printf ("<b><span foreground=\"white\">%s</span></b>", text);
+        gtk_label_set_markup (GTK_LABEL (window->priv->clock), markup);
+        g_free (markup);
+        g_free (text);
+        g_date_time_unref (dt);
+}
+
+
+static gboolean
+update_clock_timer (GSWindow *window)
+{
+        update_clock (window);
+        queue_clock_update (window);
+        return FALSE;
+}
+
+static void
+queue_clock_update (GSWindow *window)
+{
+        int timeouttime;
+        struct timeval tv;
+
+        gettimeofday (&tv, NULL);
+        timeouttime = (G_USEC_PER_SEC - tv.tv_usec) / 1000 + 1;
+
+        /* timeout of one minute if we don't care about the seconds */
+        timeouttime += 1000 * (59 - tv.tv_sec % 60);
+
+        window->priv->clock_update_id = g_timeout_add (timeouttime, (GSourceFunc)update_clock_timer, window);
+}
+
+static char *
+get_user_display_name (void)
+{
+        const char *name;
+        char       *utf8_name;
+
+        name = g_get_real_name ();
+
+        if (name == NULL || strcmp (name, "Unknown") == 0) {
+                name = g_get_user_name ();
+        }
+
+        utf8_name = NULL;
+
+        if (name != NULL) {
+                utf8_name = g_locale_to_utf8 (name, -1, NULL, NULL, NULL);
+        }
+
+        return utf8_name;
+}
+
+static void
+update_name_label (GSWindow *window)
+{
+        char *name;
+        char *markup;
+        name = get_user_display_name ();
+        markup = g_strdup_printf ("<b><span foreground=\"white\">%s</span></b>", name);
+        gtk_label_set_markup (GTK_LABEL (window->priv->name_label), markup);
+        g_free (markup);
+        g_free (name);
+}
+
+static void
+create_panel (GSWindow *window)
+{
+        GtkWidget    *left_hbox;
+        GtkWidget    *right_hbox;
+        GtkWidget    *alignment;
+        GtkWidget    *right_alignment;
+        GtkWidget    *image;
+        GtkSizeGroup *sg;
+        GdkRGBA       bg;
+        GdkRGBA       fg;
+        int           all_states;
+        GIcon        *gicon;
+
+        bg.red = 0;
+        bg.green = 0;
+        bg.blue = 0;
+        bg.alpha = 1.0;
+
+        fg.red = 1.0;
+        fg.green = 1.0;
+        fg.blue = 1.0;
+        fg.alpha = 1.0;
+
+        all_states = 0;
+
+        gtk_widget_override_background_color (window->priv->panel, all_states, &bg);
+        gtk_widget_override_color (window->priv->panel, all_states, &fg);
+        gtk_container_set_border_width (GTK_CONTAINER (window->priv->panel), 0);
+
+        g_signal_connect (window->priv->panel, "draw", G_CALLBACK (on_panel_draw), window);
+
+        left_hbox = gtk_hbox_new (FALSE, 6);
+        gtk_box_pack_start (GTK_BOX (window->priv->panel), left_hbox, TRUE, TRUE, 0);
+
+        alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
+        gtk_box_pack_start (GTK_BOX (window->priv->panel), alignment, FALSE, FALSE, 0);
+        window->priv->clock = gtk_label_new (NULL);
+        gtk_misc_set_padding (GTK_MISC (window->priv->clock), 4, 4);
+        update_clock (window);
+        queue_clock_update (window);
+        gtk_container_add (GTK_CONTAINER (alignment), window->priv->clock);
+
+        right_alignment = gtk_alignment_new (1, 0.5, 1, 1);
+        gtk_box_pack_end (GTK_BOX (window->priv->panel), right_alignment, TRUE, TRUE, 0);
+        gtk_alignment_set_padding (GTK_ALIGNMENT (right_alignment),
+                                   0, 0, 10, 10);
+
+        right_hbox = gtk_hbox_new (FALSE, 6);
+        gtk_container_add (GTK_CONTAINER (right_alignment), right_hbox);
+
+        window->priv->name_label = gtk_label_new (NULL);
+        update_name_label (window);
+        gtk_box_pack_end (GTK_BOX (right_hbox), window->priv->name_label, FALSE, FALSE, 0);
+
+        gicon = g_themed_icon_new_with_default_fallbacks ("changes-prevent-symbolic");
+        image = gtk_image_new_from_gicon (gicon, GTK_ICON_SIZE_MENU);
+        gtk_widget_override_color (image, all_states, &fg);
+        g_object_unref (gicon);
+        gtk_box_pack_end (GTK_BOX (right_hbox), image, FALSE, FALSE, 0);
+
+        sg = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
+        gtk_size_group_add_widget (sg, left_hbox);
+        gtk_size_group_add_widget (sg, right_hbox);
+
+        gtk_widget_show_all (window->priv->panel);
+}
+
 static void
 gs_window_init (GSWindow *window)
 {
@@ -2406,12 +2463,18 @@ gs_window_init (GSWindow *window)
                                | GDK_ENTER_NOTIFY_MASK
                                | GDK_LEAVE_NOTIFY_MASK);
 
-        window->priv->vbox = gtk_vbox_new (FALSE, 12);
+        window->priv->vbox = gtk_vbox_new (FALSE, 0);
         gtk_widget_show (window->priv->vbox);
         gtk_container_add (GTK_CONTAINER (window), window->priv->vbox);
 
+        window->priv->panel = gtk_hbox_new (FALSE, 12);
+        gtk_widget_show (window->priv->panel);
+        gtk_box_pack_start (GTK_BOX (window->priv->vbox), window->priv->panel, FALSE, FALSE, 0);
+        create_panel (window);
+
         window->priv->drawing_area = gtk_drawing_area_new ();
         gtk_widget_show (window->priv->drawing_area);
+        gtk_widget_set_app_paintable (window->priv->drawing_area, TRUE);
         gtk_box_pack_start (GTK_BOX (window->priv->vbox), window->priv->drawing_area, TRUE, TRUE, 0);
         create_info_bar (window);
 
@@ -2446,6 +2509,10 @@ gs_window_finalize (GObject *object)
         g_free (window->priv->logout_command);
         g_free (window->priv->keyboard_command);
 
+        if (window->priv->clock_update_id > 0) {
+                g_source_remove (window->priv->clock_update_id);
+        }
+
         if (window->priv->info_bar_timer_id > 0) {
                 g_source_remove (window->priv->info_bar_timer_id);
         }
@@ -2463,8 +2530,8 @@ gs_window_finalize (GObject *object)
 
         gs_window_dialog_finish (window);
 
-        if (window->priv->background_pixmap) {
-               g_object_unref (window->priv->background_pixmap);
+        if (window->priv->background_surface) {
+               cairo_surface_destroy (window->priv->background_surface);
         }
 
         G_OBJECT_CLASS (gs_window_parent_class)->finalize (object);
@@ -2482,6 +2549,7 @@ gs_window_new (GdkScreen *screen,
                                "screen", screen,
                                "monitor", monitor,
                                "lock-enabled", lock_enabled,
+                               "app-paintable", TRUE,
                                NULL);
 
         return GS_WINDOW (result);
